from __future__ import annotations

import base64
import asyncio
import threading
from collections.abc import Coroutine
from contextlib import AbstractContextManager
from typing import TYPE_CHECKING, Any, Callable

import viser

from . import _viser_private as impl
from ._export import ExportBuilder
from ._runtime import runtime_source
from ._runtime_messages import Viser4dRuntimeReadyMessage
from .audio import AudioApi
from ._validation import require_positive_float
from .timeline._playback import ClientPlaybackHandle
from .timeline._recording import SceneRecorder, TimelineContext
from .timeline._store import TimelineStore

if TYPE_CHECKING:
    from ._viser_private import ClientHandle


class Viser4dServer(viser.ViserServer):
    """Viser server with timestep recording, playback, and synced audio."""

    # Override the parent's plain ``scene`` attribute with a property so that
    # ``server.scene`` transparently returns the timeline-owned scene API while
    # inside a ``server.at(t)`` block.  Outside of ``at()``, it returns the
    # regular live scene.
    @property
    def scene(self) -> impl.SceneApi:
        recorder = self.__dict__.get("_recorder")
        if recorder is not None and recorder.active_step is not None:
            return recorder.scene
        return self._live_scene

    @scene.setter
    def scene(self, value: impl.SceneApi) -> None:
        self._live_scene = value

    def __init__(self, num_steps: int, fps: float = 30.0, **kwargs) -> None:
        """Initialize the timeline runtime and client playback state."""
        if num_steps < 1:
            raise ValueError(f"num_steps must be >= 1, got {num_steps}.")
        super().__init__(**kwargs)

        self.num_steps = num_steps
        self._timeline_fps = require_positive_float("fps", fps)
        self._timeline = TimelineStore(
            self.num_steps,
            flush_executor=impl.server_thread_executor(self),
        )
        self._client_playbacks: dict[int, ClientPlaybackHandle] = {}
        self._client_playbacks_lock = threading.Lock()
        self._timestep_callbacks: list[
            Callable[[ClientHandle, int], None | Coroutine[Any, Any, None]]
        ] = []
        self._playback_callbacks: list[
            Callable[[ClientHandle, bool], None | Coroutine[Any, Any, None]]
        ] = []
        self._stop_event = threading.Event()
        self._recorder = SceneRecorder(self, self._timeline)
        self._export_builder = ExportBuilder(self, self._timeline)
        self.audio = AudioApi(self)

        # Load the browser runtime once so live clients can handle timeline/audio messages.
        impl.queue_server_message(self, impl.run_javascript_message(runtime_source()))
        impl.register_message_handler(
            self,
            Viser4dRuntimeReadyMessage,
            self._handle_runtime_ready,
        )

        @self.on_client_connect
        def _attach_playback(client: ClientHandle) -> None:
            playback = ClientPlaybackHandle(
                self,
                client,
                brand_color=impl.playback_brand_color(self),
            )
            with self._client_playbacks_lock:
                self._client_playbacks[client.client_id] = playback

        @self.on_client_disconnect
        def _detach_playback(client: ClientHandle) -> None:
            with self._client_playbacks_lock:
                self._client_playbacks.pop(client.client_id, None)

    def at(self, t: int) -> AbstractContextManager[TimelineContext]:
        """Expose the timeline APIs for timestep ``t``.

        Existing code can keep using ``server.scene`` inside the context. The
        returned ``timeline`` object provides the same timeline-owned scene and
        audio APIs explicitly.
        """
        return self._recorder.at(t)

    @property
    def fps(self) -> float:
        """Timeline step rate used for recording, audio timing, and export."""
        return self._timeline_fps

    def play(self, speed: float | None = None, loop: bool | None = None) -> None:
        """Ask connected clients to play from their own current timesteps."""
        next_speed = None
        if speed is not None:
            next_speed = require_positive_float("speed", speed)
        for playback in self._client_playback_values():
            playback.play(speed=next_speed, loop=loop)

    def pause(self) -> None:
        """Ask connected clients to pause at their current timesteps."""
        for playback in self._client_playback_values():
            playback.pause()

    def refresh(self) -> None:
        """Redraw the current timestep on all connected clients."""
        for playback in self._client_playback_values():
            playback.refresh()

    def set_playback_speed(self, speed: float) -> None:
        """Update connected client playback speed without starting playback."""
        next_speed = require_positive_float("speed", speed)
        for playback in self._client_playback_values():
            playback.set_speed(next_speed)

    def on_timestep_change(
        self,
        callback: Callable[[ClientHandle, int], None | Coroutine[Any, Any, None]],
    ) -> None:
        """Register a callback for any committed client timestep change."""
        self._timestep_callbacks.append(callback)

    def on_playback_change(
        self,
        callback: Callable[[ClientHandle, bool], None | Coroutine[Any, Any, None]],
    ) -> None:
        """Register a callback for client play/pause state changes."""
        self._playback_callbacks.append(callback)

    def get_client_playback(self, client_id: int) -> ClientPlaybackHandle | None:
        """Return one connected client playback handle, if present."""
        with self._client_playbacks_lock:
            return self._client_playbacks.get(client_id)

    def get_client_playbacks(self) -> dict[int, ClientPlaybackHandle]:
        """Return a copy of the connected client playback handles."""
        with self._client_playbacks_lock:
            return self._client_playbacks.copy()

    def sleep_forever(self) -> None:
        """Block until the server is stopped."""
        while not self._stop_event.wait(3600):
            pass

    def serialize(
        self,
        *,
        start_timestep: int = 0,
        end_timestep: int | None = None,
    ) -> bytes:
        """Serialize the recorded timeline to bytes."""
        return self._export_builder.serialize(
            start_timestep=start_timestep,
            end_timestep=end_timestep,
        )

    def as_html(
        self,
        *,
        dark_mode: bool = False,
        start_timestep: int = 0,
        end_timestep: int | None = None,
    ) -> str:
        """Get a self-contained HTML string for the recorded timeline."""
        scene_bytes = self.serialize(
            start_timestep=start_timestep,
            end_timestep=end_timestep,
        )
        scene_b64 = base64.b64encode(scene_bytes).decode("ascii")
        client_html = impl.viser_client_html()
        dark_mode_str = "true" if dark_mode else "false"
        inject_script = (
            f"<script>"
            f'window.__VISER_EMBED_DATA__="{scene_b64}";'
            f"window.__VISER_EMBED_CONFIG__={{darkMode:{dark_mode_str}}};"
            f"</script>"
        )
        head_end = client_html.index("</head>")
        return client_html[:head_end] + inject_script + client_html[head_end:]

    def stop(self) -> None:
        """Shut down the underlying viser server."""
        self._stop_event.set()
        self._recorder.close()
        self._timeline.close()
        super().stop()

    def _dispatch_audio_update(self, message: impl.Message) -> None:
        self._recorder.dispatch_audio_update(message)

    def _client_playback_values(self) -> list[ClientPlaybackHandle]:
        with self._client_playbacks_lock:
            return list(self._client_playbacks.values())

    def _handle_runtime_ready(
        self, client_id: int, _message: Viser4dRuntimeReadyMessage
    ) -> None:
        playback = self.get_client_playback(client_id)
        if playback is not None:
            playback.mark_runtime_ready()

    def _dispatch_timestep_change(self, client: ClientHandle, timestep: int) -> None:
        for callback in list(self._timestep_callbacks):
            result = callback(client, timestep)
            if asyncio.iscoroutine(result):
                self.get_event_loop().create_task(result)

    def _dispatch_playback_change(self, client: ClientHandle, is_playing: bool) -> None:
        for callback in list(self._playback_callbacks):
            result = callback(client, is_playing)
            if asyncio.iscoroutine(result):
                self.get_event_loop().create_task(result)



"""Checkpoint state model and all logic for building, applying, and persisting checkpoints."""

from __future__ import annotations

import base64
from dataclasses import dataclass, field
from pathlib import Path

import msgspec
import numpy as np
import zstandard

from ..audio._api import audio_array_payload
from ..audio._messages import AddAudioMessage
from .._types import StoredMessage
from ._messages_util import (
    TimelineStep,
    extract_message_name,
    store_raw_message,
    stored_dict,
    stored_float,
    stored_int,
)


@dataclass
class AudioTrackState:
    sample_rate: int
    waveform: np.ndarray
    volume: float = 1.0


@dataclass
class CheckpointState:
    scene_updates: dict[str, StoredMessage] = field(default_factory=dict)
    key_to_node: dict[str, str | None] = field(default_factory=dict)
    audio_tracks: dict[str, AudioTrackState] = field(default_factory=dict)


class _CheckpointFilePayload(msgspec.Struct):
    sceneUpdates: list[tuple[str, StoredMessage]]
    keyToNode: list[tuple[str, str | None]]
    audioTracks: list["_CheckpointAudioTrackPayload"]


class _CheckpointAudioTrackPayload(msgspec.Struct):
    name: str
    sampleRate: int
    waveform: bytes
    dtype: str
    shape: list[int]
    volume: float


def apply_scene_message(
    state: CheckpointState, key: str, message: StoredMessage
) -> None:
    name = extract_message_name(message)
    if message.payload.get("type") == "RemoveSceneNodeMessage" and isinstance(
        name, str
    ):
        prefix = f"{name}/"
        stale_keys = [
            k
            for k, node in state.key_to_node.items()
            if node == name or (isinstance(node, str) and node.startswith(prefix))
        ]
        for k in stale_keys:
            del state.scene_updates[k]
            del state.key_to_node[k]
        return
    state.scene_updates.pop(key, None)
    state.scene_updates[key] = message
    state.key_to_node[key] = name


def apply_audio_message(state: CheckpointState, message: StoredMessage) -> None:
    message_type = message.payload.get("type")
    name = extract_message_name(message)
    if not isinstance(message_type, str) or not isinstance(name, str):
        return
    if message_type == "RemoveAudioMessage":
        state.audio_tracks.pop(name, None)
        return
    if message_type == "AddAudioMessage":
        state.audio_tracks[name] = AudioTrackState(
            sample_rate=stored_int(message.payload["sampleRate"]),
            waveform=_decode_audio_payload(stored_dict(message.payload["waveform"])),
            volume=stored_float(message.payload["volume"]),
        )
        return
    track = state.audio_tracks.get(name)
    if track is None:
        return
    if message_type == "SetAudioVolumeMessage":
        track.volume = stored_float(message.payload["volume"])
    elif message_type == "SetAudioWaveformMessage":
        track.waveform = _decode_audio_payload(stored_dict(message.payload["waveform"]))
    elif message_type == "AppendAudioMessage":
        track.waveform = _append_audio_waveform(
            track.waveform,
            _decode_audio_payload(stored_dict(message.payload["waveform"])),
        )


def apply_steps(state: CheckpointState, steps: list[TimelineStep]) -> None:
    """Apply all scene and audio updates from a sequence of steps."""
    for step in steps:
        for key, message in step.scene_updates.items():
            apply_scene_message(state, key, message)
        for message in step.audio_updates:
            apply_audio_message(state, message)


def copy_checkpoint(state: CheckpointState) -> CheckpointState:
    return CheckpointState(
        scene_updates=dict(state.scene_updates),
        key_to_node=dict(state.key_to_node),
        audio_tracks={
            name: AudioTrackState(
                sample_rate=track.sample_rate,
                waveform=track.waveform.copy(),
                volume=track.volume,
            )
            for name, track in state.audio_tracks.items()
        },
    )


def checkpoint_messages(state: CheckpointState) -> list[StoredMessage]:
    """Build the ordered message list that reconstructs this checkpoint's state."""
    unnamed_messages = [
        message
        for key, message in state.scene_updates.items()
        if state.key_to_node.get(key) is None
    ]
    node_messages: dict[str, list[StoredMessage]] = {}
    for key, message in state.scene_updates.items():
        name = state.key_to_node.get(key)
        if name is None:
            continue
        node_messages.setdefault(name, []).append(message)
    scene_messages = list(unnamed_messages)
    for name in sorted(node_messages, key=_scene_node_sort_key):
        messages = node_messages[name]
        create = [m for m in messages if _is_create_scene_message(m)]
        scene_messages.extend(create)
        scene_messages.extend(m for m in messages if not _is_create_scene_message(m))
    audio_messages = [
        store_raw_message(
            AddAudioMessage(
                name=name,
                sampleRate=track.sample_rate,
                waveform=audio_array_payload(track.waveform),
                volume=track.volume,
            )
        )
        for name, track in sorted(state.audio_tracks.items())
    ]
    return scene_messages + audio_messages


def write_checkpoint_file(path: Path, state: CheckpointState) -> None:
    audio_tracks = [
        _CheckpointAudioTrackPayload(
            name=name,
            sampleRate=track.sample_rate,
            waveform=track.waveform.tobytes(),
            dtype=track.waveform.dtype.str,
            shape=list(track.waveform.shape),
            volume=track.volume,
        )
        for name, track in sorted(state.audio_tracks.items())
    ]
    payload = _CheckpointFilePayload(
        sceneUpdates=list(state.scene_updates.items()),
        keyToNode=list(state.key_to_node.items()),
        audioTracks=audio_tracks,
    )
    packed = msgspec.msgpack.encode(payload)
    compressed = zstandard.ZstdCompressor(level=6).compress(packed)
    path.write_bytes(compressed)


def load_checkpoint_file(path: Path) -> CheckpointState:
    raw = zstandard.ZstdDecompressor().decompress(path.read_bytes())
    payload = msgspec.msgpack.decode(raw, type=_CheckpointFilePayload)
    scene_updates = dict(payload.sceneUpdates)
    key_to_node = dict(payload.keyToNode)
    audio_tracks: dict[str, AudioTrackState] = {}
    for td in payload.audioTracks:
        shape = tuple(td.shape)
        arr = np.frombuffer(td.waveform, dtype=np.dtype(td.dtype))
        audio_tracks[td.name] = AudioTrackState(
            sample_rate=td.sampleRate,
            waveform=np.ascontiguousarray(arr.reshape(shape)),
            volume=td.volume,
        )
    return CheckpointState(
        scene_updates=scene_updates,
        key_to_node=key_to_node,
        audio_tracks=audio_tracks,
    )


def _scene_node_sort_key(name: str) -> tuple[int, str]:
    return (name.count("/"), name)


def _is_create_scene_message(message: StoredMessage) -> bool:
    return "props" in message.payload


def _decode_audio_payload(payload: dict[str, object]) -> np.ndarray:
    dtype = np.dtype(str(payload["dtype"]))
    num_channels = stored_int(payload["numChannels"])
    num_frames = stored_int(payload["numFrames"])
    data = np.frombuffer(base64.b64decode(str(payload["data"])), dtype=dtype)
    if num_channels <= 1:
        return np.ascontiguousarray(data[:num_frames])
    return np.ascontiguousarray(data.reshape(num_frames, num_channels))


def _append_audio_waveform(head: np.ndarray, tail: np.ndarray) -> np.ndarray:
    if head.ndim != tail.ndim:
        raise ValueError("Audio append must preserve waveform dimensionality.")
    return np.ascontiguousarray(np.concatenate((head, tail), axis=0))



import threading
import time
from typing import cast

import msgspec
import numpy as np
import pytest
import zstandard

import viser4d


def _deserialize_recording(blob: bytes) -> dict[str, object]:
    hybrid_size = int.from_bytes(blob[:8], "little")
    hybrid = zstandard.ZstdDecompressor().decompress(
        blob[8:], max_output_size=hybrid_size
    )
    msgpack_size = int.from_bytes(hybrid[:8], "little")
    return cast(dict[str, object], msgspec.msgpack.decode(hybrid[8 : 8 + msgpack_size]))


def test_audio_requires_timestep_context() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with pytest.raises(RuntimeError):
            server.audio.add_track(
                "/audio", data=np.zeros(4, dtype=np.int16), sample_rate=8_000
            )
    finally:
        server.stop()


def test_num_steps_must_be_positive() -> None:
    with pytest.raises(ValueError, match="num_steps must be >= 1"):
        viser4d.Viser4dServer(num_steps=0, port=0, verbose=False)


def test_fps_and_speed_must_be_positive() -> None:
    with pytest.raises(ValueError, match="fps must be a positive finite float"):
        viser4d.Viser4dServer(num_steps=1, fps=0.0, port=0, verbose=False)

    server = viser4d.Viser4dServer(num_steps=3, port=0, verbose=False)
    try:
        with pytest.raises(ValueError, match="speed must be a positive finite float"):
            server.play(speed=0.0)
        with pytest.raises(ValueError, match="speed must be a positive finite float"):
            server.set_playback_speed(-1.0)
    finally:
        server.stop()


def test_timeline_operations_serialize_and_playback_commands() -> None:
    server = viser4d.Viser4dServer(num_steps=3, port=0, verbose=False)
    try:
        with server.at(0) as timeline:
            frame = timeline.scene.add_frame("/frame")
            audio = timeline.audio.add_track(
                "/audio",
                data=np.array([0, 1, 2], dtype=np.int16),
                sample_rate=16_000,
            )
        with server.at(1):
            frame.position = (1.0, 2.0, 3.0)
            audio.volume = 0.25

        server.set_playback_speed(0.8)
        server.play()
        server.pause()

        assert server.serialize()
        assert server.serialize(start_timestep=1, end_timestep=1)
    finally:
        server.stop()


def test_at_preserves_server_scene_backwards_compatibility() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with server.at(0):
            joint = server.scene.add_frame("/joint")
        server.scene.add_frame("/static")

        joint.position = (2.0, 0.0, 0.0)

        recording = _deserialize_recording(server.serialize())
        messages = cast(list[tuple[float, dict[str, object]]], recording["messages"])
        creation_times = [
            time
            for time, message in messages
            if message.get("type") == "FrameMessage" and message.get("name") == "/joint"
        ]
        positions = [
            tuple(cast(list[float], message["position"]))
            for _, message in messages
            if message.get("type") == "SetPositionMessage"
            and message.get("name") == "/joint"
        ]
        static_times = [
            time
            for time, message in messages
            if message.get("type") == "FrameMessage"
            and message.get("name") == "/static"
        ]

        assert creation_times == [0.0]
        assert positions == [(2.0, 0.0, 0.0)]
        assert static_times == [0.0]
    finally:
        server.stop()


def test_same_step_scene_updates_serialize_latest_value_once() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with server.at(0) as timeline:
            joint = timeline.scene.add_frame("/joint")
        with server.at(0):
            joint.position = (1.0, 0.0, 0.0)
        with server.at(0):
            joint.position = (2.0, 0.0, 0.0)

        recording = _deserialize_recording(server.serialize())
        messages = cast(list[tuple[float, dict[str, object]]], recording["messages"])
        positions = [
            tuple(cast(list[float], message["position"]))
            for _, message in messages
            if message.get("type") == "SetPositionMessage"
            and message.get("name") == "/joint"
        ]

        assert positions == [(2.0, 0.0, 0.0)]
    finally:
        server.stop()


def test_serialize_rejects_invalid_timestep_range() -> None:
    server = viser4d.Viser4dServer(num_steps=3, port=0, verbose=False)
    try:
        with pytest.raises(
            ValueError, match="start_timestep must be less than or equal"
        ):
            server.serialize(start_timestep=2, end_timestep=1)
    finally:
        server.stop()


def test_at_rejects_static_name_collisions() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        server.scene.add_icosphere("/joint", position=(0.0, 0.0, 0.0))
        with pytest.raises(RuntimeError, match="static scene node"):
            with server.at(0) as timeline:
                timeline.scene.add_icosphere("/joint", position=(1.0, 0.0, 0.0))
    finally:
        server.stop()


def test_at_allows_recreating_timeline_nodes() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with server.at(0) as timeline:
            timeline.scene.add_icosphere("/joint", position=(0.0, 0.0, 0.0))

        with server.at(1) as timeline:
            joint = timeline.scene.add_icosphere("/joint", position=(1.0, 0.0, 0.0))

        joint.position = (2.0, 0.0, 0.0)

        recording = _deserialize_recording(server.serialize())
        messages = cast(list[tuple[float, dict[str, object]]], recording["messages"])
        creation_times = [
            time
            for time, message in messages
            if message.get("type") == "IcosphereMessage"
            and message.get("name") == "/joint"
        ]
        positions = [
            (time, tuple(cast(list[float], message["position"])))
            for time, message in messages
            if message.get("type") == "SetPositionMessage"
            and message.get("name") == "/joint"
        ]

        assert creation_times == [0.0, 1.0 / server.fps]
        assert positions == [(1.0 / server.fps, (2.0, 0.0, 0.0))]
    finally:
        server.stop()


def test_stop_unblocks_sleep_forever() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    sleeper = threading.Thread(target=server.sleep_forever)
    sleeper.start()

    try:
        time.sleep(0.05)
        server.stop()
        sleeper.join(timeout=1.0)
        assert sleeper.is_alive() is False
    finally:
        if sleeper.is_alive():
            server.stop()
            sleeper.join(timeout=1.0)


def test_audio_waveform_reflects_appended_data() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with server.at(0) as timeline:
            audio = timeline.audio.add_track(
                "/audio",
                data=np.array([1, 2], dtype=np.int16),
                sample_rate=16_000,
            )

        audio.append(np.array([3, 4], dtype=np.int16))
        audio.append(np.array([5, 6], dtype=np.int16))

        expected = np.array([1, 2, 3, 4, 5, 6], dtype=np.int16)
        assert np.array_equal(audio.waveform, expected)
    finally:
        server.stop()


def test_audio_rejects_non_mono_or_stereo_shapes() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with server.at(0) as timeline:
            with pytest.raises(ValueError, match="mono or stereo"):
                timeline.audio.add_track(
                    "/audio",
                    data=np.zeros((2, 2, 2), dtype=np.float32),
                    sample_rate=16_000,
                )
            with pytest.raises(ValueError, match="mono or stereo"):
                timeline.audio.add_track(
                    "/audio-3ch",
                    data=np.zeros((4, 3), dtype=np.float32),
                    sample_rate=16_000,
                )
    finally:
        server.stop()


def test_stereo_audio_append_preserves_channel_layout() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with server.at(0) as timeline:
            audio = timeline.audio.add_track(
                "/audio",
                data=np.array([[1, 10], [2, 20]], dtype=np.int16),
                sample_rate=16_000,
            )

        audio.append(np.array([[3, 30], [4, 40]], dtype=np.int16))

        expected = np.array([[1, 10], [2, 20], [3, 30], [4, 40]], dtype=np.int16)
        assert np.array_equal(audio.waveform, expected)

        with pytest.raises(ValueError, match="channel count"):
            audio.append(np.array([5, 6], dtype=np.int16))
    finally:
        server.stop()


def test_same_step_audio_events_serialize_without_deduping() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with server.at(0) as timeline:
            audio = timeline.audio.add_track(
                "/audio",
                data=np.array([1, 2], dtype=np.int16),
                sample_rate=16_000,
            )

        with server.at(1):
            audio.append(np.array([3, 4], dtype=np.int16))
            audio.append(np.array([5, 6], dtype=np.int16))

        recording = _deserialize_recording(server.serialize())
        messages = cast(list[tuple[float, dict[str, object]]], recording["messages"])
        append_messages = [
            message
            for _, message in messages
            if message.get("type") == "AppendAudioMessage"
            and message.get("name") == "/audio"
        ]

        assert len(append_messages) == 2
    finally:
        server.stop()


def test_post_recording_timeline_updates_persist_in_serialization() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with server.at(0) as timeline:
            joint = timeline.scene.add_frame("/joint")

        joint.position = (2.0, 0.0, 0.0)

        recording = _deserialize_recording(server.serialize())
        messages = cast(list[tuple[float, dict[str, object]]], recording["messages"])
        positions = [
            tuple(cast(list[float], message["position"]))
            for _, message in messages
            if message.get("type") == "SetPositionMessage"
            and message.get("name") == "/joint"
        ]

        assert positions == [(2.0, 0.0, 0.0)]
    finally:
        server.stop()


def test_late_created_timeline_nodes_keep_their_creation_step() -> None:
    server = viser4d.Viser4dServer(num_steps=6, fps=1.0, port=0, verbose=False)
    try:
        with server.at(5) as timeline:
            joint = timeline.scene.add_frame("/joint")

        joint.position = (2.0, 0.0, 0.0)

        recording = _deserialize_recording(server.serialize())
        messages = cast(list[tuple[float, dict[str, object]]], recording["messages"])

        creation_times = [
            time
            for time, message in messages
            if message.get("type") == "FrameMessage" and message.get("name") == "/joint"
        ]
        position_times = [
            time
            for time, message in messages
            if message.get("type") == "SetPositionMessage"
            and message.get("name") == "/joint"
        ]

        assert creation_times == [5.0]
        assert position_times == [5.0]
    finally:
        server.stop()


def test_serialization_survives_block_eviction_to_disk() -> None:
    server = viser4d.Viser4dServer(num_steps=300, fps=1.0, port=0, verbose=False)
    try:
        with server.at(0) as timeline:
            joint = timeline.scene.add_frame("/joint")

        for step in (1, 64, 128, 192, 256):
            with server.at(step):
                joint.position = (float(step), 0.0, 0.0)

        recording = _deserialize_recording(server.serialize())
        messages = cast(list[tuple[float, dict[str, object]]], recording["messages"])
        position_times = [
            time
            for time, message in messages
            if message.get("type") == "SetPositionMessage"
            and message.get("name") == "/joint"
        ]
        positions = [
            tuple(cast(list[float], message["position"]))
            for _, message in messages
            if message.get("type") == "SetPositionMessage"
            and message.get("name") == "/joint"
        ]

        assert position_times == [1.0, 64.0, 128.0, 192.0, 256.0]
        assert positions == [
            (1.0, 0.0, 0.0),
            (64.0, 0.0, 0.0),
            (128.0, 0.0, 0.0),
            (192.0, 0.0, 0.0),
            (256.0, 0.0, 0.0),
        ]
    finally:
        server.stop()



"""Block-backed timeline storage with LRU caching and checkpoint management."""

from __future__ import annotations

import math
import tempfile
import threading
from collections import OrderedDict
from concurrent.futures import Executor, Future
from dataclasses import dataclass
from pathlib import Path

import msgspec
import zstandard

from .. import _viser_private as impl
from ..audio._messages import is_audio_message
from .._types import RuntimeBlockPayload, StoredMessage
from ._checkpoint import (
    CheckpointState,
    apply_steps,
    checkpoint_messages,
    copy_checkpoint,
    load_checkpoint_file,
    write_checkpoint_file,
)
from ._messages_util import (
    TimelineStep,
    extract_message_name,
    is_scene_message,
    store_raw_message,
)


@dataclass
class TimelineBlock:
    steps: list[TimelineStep]
    dirty: bool = False


class _BlockFilePayload(msgspec.Struct):
    sceneSteps: list[list[tuple[str, StoredMessage]]]
    audioSteps: list[list[StoredMessage]]


class TimelineStore:
    """Block-backed storage for timeline-owned steps and live scene updates."""

    def __init__(
        self,
        num_steps: int,
        *,
        block_size: int = 64,
        max_loaded_blocks: int = 4,
        max_cached_checkpoints: int = 4,
        flush_executor: Executor,
    ) -> None:
        if block_size < 1:
            raise ValueError(f"block_size must be >= 1, got {block_size}.")
        if max_loaded_blocks < 1:
            raise ValueError(
                f"max_loaded_blocks must be >= 1, got {max_loaded_blocks}."
            )
        if max_cached_checkpoints < 1:
            raise ValueError(
                f"max_cached_checkpoints must be >= 1, got {max_cached_checkpoints}."
            )
        self.num_steps = num_steps
        self.block_size = block_size
        self.block_count = math.ceil(num_steps / block_size)
        self._node_start_steps: dict[str, int] = {}
        self._last_scene_steps: dict[str, int] = {}
        self._lock = threading.RLock()
        self._loaded_blocks: OrderedDict[int, TimelineBlock] = OrderedDict()
        self._max_loaded_blocks = max_loaded_blocks
        self._checkpoint_cache: OrderedDict[int, CheckpointState] = OrderedDict()
        self._max_cached_checkpoints = max_cached_checkpoints
        self._flush_executor = flush_executor
        self._pending_flushes: dict[int, Future[None]] = {}
        self._block_dir_root = tempfile.TemporaryDirectory(prefix="viser4d-timeline-")
        self._block_dir = Path(self._block_dir_root.name)
        # Running checkpoint for eager incremental persistence during sequential recording.
        self._eager_checkpoint = CheckpointState()
        self._eager_checkpoint_next_block = 0

    def validate_step(self, step: int) -> int:
        """Return ``step`` if it is in range, else raise ``IndexError``."""
        if step < 0 or step >= self.num_steps:
            raise IndexError(
                f"Timestep {step} is out of range for {self.num_steps} steps."
            )
        return step

    def step(self, step: int) -> TimelineStep:
        """Return the storage bucket for one timestep."""
        with self._lock:
            step = self.validate_step(step)
            block = self._load_block(step // self.block_size)
            return block.steps[step % self.block_size]

    def has_node(self, name: str) -> bool:
        with self._lock:
            return name in self._node_start_steps

    def record_step(self, step: int, messages: list[impl.Message]) -> None:
        """Store one timestep's scene and audio updates."""
        with self._lock:
            step = self.validate_step(step)
            block_index = step // self.block_size
            block = self._load_block(block_index)
            step_state = block.steps[step % self.block_size]
            for message in messages:
                stored_message = store_raw_message(message)
                key = message.redundancy_key()
                name = extract_message_name(stored_message)
                if is_scene_message(stored_message):
                    if name is not None and impl.is_create_scene_node_message(message):
                        self._node_start_steps[name] = step
                    self._last_scene_steps[key] = step
                    step_state.scene_updates.pop(key, None)
                    step_state.scene_updates[key] = stored_message
                    continue
                if is_audio_message(message):
                    step_state.audio_updates.append(stored_message)
            block.dirty = True
            self._invalidate_checkpoints_after_block(block_index)

    def record_live_scene_update(self, message: impl.Message) -> int:
        with self._lock:
            stored_message = store_raw_message(message)
            name = extract_message_name(stored_message)
            key = message.redundancy_key()
            start_step = 0 if name is None else self._node_start_steps.get(name, 0)
            last_scene_step = self._last_scene_steps.get(key)
            if last_scene_step is not None:
                start_step = max(start_step, last_scene_step)
            block_index = start_step // self.block_size
            block = self._load_block(block_index)
            step_state = block.steps[start_step % self.block_size]
            step_state.scene_updates.pop(key, None)
            step_state.scene_updates[key] = stored_message
            self._last_scene_steps[key] = start_step
            block.dirty = True
            self._invalidate_checkpoints_after_block(block_index)
            return start_step

    def block_index_for_step(self, step: int) -> int:
        return self.validate_step(step) // self.block_size

    def block_payload(self, block_index: int) -> RuntimeBlockPayload:
        with self._lock:
            block_index = self._validate_block_index(block_index)
            ckpt = self._checkpoint_for_block(block_index)
            block = self._load_block(block_index)
            step_messages = [
                list(step.scene_updates.values()) + list(step.audio_updates)
                for step in block.steps
            ]
        ckpt_messages = checkpoint_messages(ckpt)
        return {
            "block": block_index,
            "checkpointMessages": ckpt_messages,
            "stepMessages": step_messages,
        }

    def close(self) -> None:
        with self._lock:
            try:
                for block_index, block in list(self._loaded_blocks.items()):
                    self._flush_block(block_index, block)
                self._loaded_blocks.clear()
                for block_index in tuple(self._pending_flushes):
                    self._wait_for_pending_flush(block_index)
            finally:
                self._block_dir_root.cleanup()

    def _block_path(self, block_index: int) -> Path:
        return self._block_dir / f"{block_index:08d}.msgpack.zst"

    def _checkpoint_path(self, block_index: int) -> Path:
        return self._block_dir / f"checkpoint_{block_index:08d}.msgpack.zst"

    def _validate_block_index(self, block_index: int) -> int:
        if block_index < 0 or block_index >= self.block_count:
            raise IndexError(
                f"Block {block_index} is out of range for {self.block_count} blocks."
            )
        return block_index

    def _load_block(self, block_index: int) -> TimelineBlock:
        block_index = self._validate_block_index(block_index)
        block = self._loaded_blocks.get(block_index)
        if block is not None:
            self._loaded_blocks.move_to_end(block_index)
            return block
        self._wait_for_pending_flush(block_index)
        path = self._block_path(block_index)
        if path.exists():
            raw = zstandard.ZstdDecompressor().decompress(path.read_bytes())
            payload = msgspec.msgpack.decode(raw, type=_BlockFilePayload)
            block = TimelineBlock(
                steps=[
                    TimelineStep(
                        scene_updates=dict(scene_updates),
                        audio_updates=audio_updates,
                    )
                    for scene_updates, audio_updates in zip(
                        payload.sceneSteps,
                        payload.audioSteps,
                        strict=True,
                    )
                ],
                dirty=False,
            )
        else:
            step_count = min(
                self.block_size, self.num_steps - block_index * self.block_size
            )
            block = TimelineBlock(steps=[TimelineStep() for _ in range(step_count)])
        self._loaded_blocks[block_index] = block
        self._evict_loaded_blocks()
        return block

    def _evict_loaded_blocks(self) -> None:
        while len(self._loaded_blocks) > self._max_loaded_blocks:
            block_index, block = self._loaded_blocks.popitem(last=False)
            self._flush_block(block_index, block)

    def _flush_block(self, block_index: int, block: TimelineBlock) -> None:
        if not block.dirty:
            return
        block.dirty = False
        pending_flush = self._pending_flushes.get(block_index)
        # Eagerly compute the checkpoint for block_index+1 while we still have
        # the block data in hand, so seeks never need to replay prior blocks.
        # Only possible when blocks are flushed in contiguous order from 0.
        next_block = block_index + 1
        ckpt: tuple[Path, CheckpointState] | None = None
        if (
            block_index == self._eager_checkpoint_next_block
            and next_block < self.block_count
        ):
            apply_steps(self._eager_checkpoint, block.steps)
            self._eager_checkpoint_next_block = next_block
            ckpt = (
                self._checkpoint_path(next_block),
                copy_checkpoint(self._eager_checkpoint),
            )
        self._pending_flushes[block_index] = self._flush_executor.submit(
            _write_block_and_checkpoint_after,
            pending_flush,
            self._block_path(block_index),
            block,
            ckpt,
        )

    def _wait_for_pending_flush(self, block_index: int) -> None:
        future = self._pending_flushes.pop(block_index, None)
        if future is not None:
            future.result()

    def _invalidate_checkpoints_after_block(self, block_index: int) -> None:
        stale = [i for i in self._checkpoint_cache if i > block_index]
        for i in stale:
            del self._checkpoint_cache[i]

    def _checkpoint_for_block(self, block_index: int) -> CheckpointState:
        block_index = self._validate_block_index(block_index)
        cached = self._checkpoint_cache.get(block_index)
        if cached is not None:
            self._checkpoint_cache.move_to_end(block_index)
            return copy_checkpoint(cached)
        # Check disk before falling back to block replay.
        disk_state = self._load_checkpoint_from_disk(block_index)
        if disk_state is not None:
            self._cache_checkpoint(block_index, disk_state)
            return disk_state
        prior_blocks = [i for i in self._checkpoint_cache if i < block_index]
        base_index = max(prior_blocks, default=0)
        state = (
            copy_checkpoint(self._checkpoint_cache[base_index])
            if base_index in self._checkpoint_cache
            else CheckpointState()
        )
        for index in range(base_index, block_index):
            block = self._load_block(index)
            apply_steps(state, block.steps)
            self._cache_checkpoint(index + 1, state)
        return copy_checkpoint(state)

    def _load_checkpoint_from_disk(self, block_index: int) -> CheckpointState | None:
        # Wait for any in-flight flush that might be writing this checkpoint.
        if block_index > 0:
            pending = self._pending_flushes.get(block_index - 1)
            if pending is not None:
                pending.result()
        path = self._checkpoint_path(block_index)
        return load_checkpoint_file(path) if path.exists() else None

    def _cache_checkpoint(self, block_index: int, state: CheckpointState) -> None:
        self._checkpoint_cache[block_index] = copy_checkpoint(state)
        self._checkpoint_cache.move_to_end(block_index)
        while len(self._checkpoint_cache) > self._max_cached_checkpoints:
            self._checkpoint_cache.popitem(last=False)


def _write_block_file(path: Path, block: TimelineBlock) -> None:
    payload = _BlockFilePayload(
        sceneSteps=[list(step.scene_updates.items()) for step in block.steps],
        audioSteps=[step.audio_updates for step in block.steps],
    )
    packed = msgspec.msgpack.encode(payload)
    compressed = zstandard.ZstdCompressor(level=6).compress(packed)
    path.write_bytes(compressed)


def _write_block_and_checkpoint_after(
    previous_flush: Future[None] | None,
    block_path: Path,
    block: TimelineBlock,
    checkpoint: tuple[Path, CheckpointState] | None,
) -> None:
    if previous_flush is not None:
        previous_flush.result()
    _write_block_file(block_path, block)
    if checkpoint is not None:
        write_checkpoint_file(*checkpoint)



"""Pure message conversion utilities for timeline storage and transport."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import cast

import viser

from .. import _viser_private as impl
from .._hybrid import serialize_zstd_hybrid_document
from .._types import StoredMessage
from ..audio._messages import is_audio_message_type


@dataclass
class TimelineStep:
    """Recorded scene and audio updates for one timestep."""

    scene_updates: dict[str, StoredMessage] = field(default_factory=dict)
    audio_updates: list[StoredMessage] = field(default_factory=list)


def store_raw_message(message: impl.Message) -> StoredMessage:
    """Capture one viser message in placeholder-plus-buffer form."""
    buffers: list[memoryview] = []
    payload = cast(
        dict[str, object],
        message.as_serializable_dict(binary_buffers=buffers),
    )
    return StoredMessage(payload, tuple(bytes(buffer) for buffer in buffers))


def store_raw_messages(messages: list[impl.Message]) -> list[StoredMessage]:
    return [store_raw_message(message) for message in messages]


def stored_int(value: object) -> int:
    if isinstance(value, bool) or not isinstance(value, (int, float, str)):
        raise TypeError(f"Expected int-like stored value, got {type(value).__name__}.")
    return int(value)


def stored_float(value: object) -> float:
    if isinstance(value, bool) or not isinstance(value, (int, float, str)):
        raise TypeError(
            f"Expected float-like stored value, got {type(value).__name__}."
        )
    return float(value)


def stored_dict(value: object) -> dict[str, object]:
    if not isinstance(value, dict):
        raise TypeError(f"Expected dict stored value, got {type(value).__name__}.")
    return cast(dict[str, object], value)


def extract_message_name(message: StoredMessage) -> str | None:
    name = message.payload.get("name")
    return name if isinstance(name, str) and name else None


def is_scene_message(message: StoredMessage) -> bool:
    message_type = message.payload.get("type")
    return (
        isinstance(message_type, str)
        and not message_type.startswith("Gui")
        and not is_audio_message_type(message_type)
    )


def serialize_viser_recording(
    messages: list[tuple[float, StoredMessage]],
    *,
    duration_seconds: float = 0.0,
) -> bytes:
    return serialize_zstd_hybrid_document(
        {
            "durationSeconds": duration_seconds,
            "messages": messages,
            "viserVersion": viser.__version__,
        },
        level=12,
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeAlias, TypedDict


StoredPayload: TypeAlias = dict[str, object]


@dataclass(frozen=True)
class StoredMessage:
    payload: StoredPayload
    buffers: tuple[bytes, ...] = ()


class AudioArrayPayload(TypedDict):
    dtype: str
    numChannels: int
    numFrames: int
    data: str


class RuntimeConfig(TypedDict):
    numSteps: int
    blockSize: int
    timelineFps: float
    speed: float
    loop: bool


class ClientRuntimeConfig(RuntimeConfig):
    blockRequestSyncUuid: str
    timelineSliderUuid: str
    speedSliderUuid: str
    stepButtonsUuid: str
    playButtonUuid: str
    pauseButtonUuid: str
    speedSyncUuid: str
    playbackStateSyncUuid: str
    timestepSyncUuid: str


class RuntimeBlockPayload(TypedDict):
    block: int
    checkpointMessages: list[StoredMessage]
    stepMessages: list[list[StoredMessage]]



from __future__ import annotations

import pathlib

from . import _client_autobuild


RUNTIME_MARKER = "/*__VISER4D_RUNTIME__*/"


def runtime_source() -> str:
    runtime_path = pathlib.Path(__file__).resolve().parent / "runtime.js"
    _client_autobuild.ensure_client_is_built()
    try:
        return RUNTIME_MARKER + runtime_path.read_text()
    except FileNotFoundError as exc:
        raise RuntimeError(
            "Missing generated client bundle at src/viser4d/runtime.js. "
            "viser4d could not rebuild it automatically."
        ) from exc



from __future__ import annotations

from typing import Any, cast

import msgspec
import numpy as np
import zstandard

from ._types import StoredMessage

_BINARY_INDEX_KEY = "__binary_index"
_BINARY_LENGTHS_KEY = "binaryBufferLengths"
_DTYPE_KEY = "dtype"


def _byte_view(value: memoryview | bytes | bytearray) -> memoryview:
    view = value if isinstance(value, memoryview) else memoryview(value)
    if not view.c_contiguous:
        view = memoryview(view.tobytes())
    if view.format != "B" or view.itemsize != 1:
        view = view.cast("B")
    return view


def _append_placeholder(
    buffers: list[memoryview],
    data: memoryview | bytes | bytearray,
    *,
    dtype: str,
) -> dict[str, object]:
    byte_view = _byte_view(data)
    buffers.append(byte_view)
    return {
        _BINARY_INDEX_KEY: len(buffers) - 1,
        _DTYPE_KEY: dtype,
    }


def _remap_placeholders(
    value: Any,
    *,
    buffer_offset: int,
    buffer_count: int,
) -> Any:
    if isinstance(value, dict):
        idx = value.get(_BINARY_INDEX_KEY)
        dtype = value.get(_DTYPE_KEY)
        if isinstance(idx, int) and isinstance(dtype, str):
            if not 0 <= idx < buffer_count:
                raise ValueError(f"Binary buffer index {idx} is out of range.")
            return {
                _BINARY_INDEX_KEY: buffer_offset + idx,
                _DTYPE_KEY: dtype,
            }
        return {
            str(key): _remap_placeholders(
                inner,
                buffer_offset=buffer_offset,
                buffer_count=buffer_count,
            )
            for key, inner in value.items()
        }
    if isinstance(value, (tuple, list)):
        return [
            _remap_placeholders(
                item,
                buffer_offset=buffer_offset,
                buffer_count=buffer_count,
            )
            for item in value
        ]
    return value


def _append_stored_message(
    buffers: list[memoryview],
    message: StoredMessage,
) -> dict[str, object]:
    buffer_offset = len(buffers)
    buffers.extend(_byte_view(buffer) for buffer in message.buffers)
    return cast(
        dict[str, object],
        _remap_placeholders(
            message.payload,
            buffer_offset=buffer_offset,
            buffer_count=len(message.buffers),
        ),
    )


def _inflate_placeholders(
    value: object,
    *,
    buffers: tuple[bytes, ...],
) -> object:
    if isinstance(value, dict):
        record = cast(dict[str, object], value)
        idx = record.get(_BINARY_INDEX_KEY)
        dtype = record.get(_DTYPE_KEY)
        if isinstance(idx, int) and isinstance(dtype, str):
            if not 0 <= idx < len(buffers):
                raise ValueError(f"Binary buffer index {idx} is out of range.")
            return np.frombuffer(buffers[idx], dtype=np.dtype(dtype))
        return {
            str(key): _inflate_placeholders(inner, buffers=buffers)
            for key, inner in record.items()
        }
    if isinstance(value, list):
        return [_inflate_placeholders(item, buffers=buffers) for item in value]
    if isinstance(value, tuple):
        return tuple(_inflate_placeholders(item, buffers=buffers) for item in value)
    return value


def inflate_stored_message(message: StoredMessage) -> dict[str, object]:
    return cast(
        dict[str, object],
        _inflate_placeholders(message.payload, buffers=message.buffers),
    )


def inflate_stored_messages(messages: list[StoredMessage]) -> list[dict[str, object]]:
    return [inflate_stored_message(message) for message in messages]


def _to_hybrid_value(value: Any, buffers: list[memoryview]) -> Any:
    if isinstance(value, StoredMessage):
        return _append_stored_message(buffers, value)
    if isinstance(value, dict):
        return {
            str(key): _to_hybrid_value(inner, buffers) for key, inner in value.items()
        }
    if isinstance(value, np.ndarray):
        data = value.data if value.data.c_contiguous else value.copy().data
        return _append_placeholder(
            buffers, cast(memoryview, data), dtype=value.dtype.str
        )
    if isinstance(value, (memoryview, bytes, bytearray)):
        return _append_placeholder(buffers, value, dtype="|u1")
    if isinstance(value, (tuple, list)):
        return [_to_hybrid_value(item, buffers) for item in value]
    return value


def encode_hybrid_document(document: object) -> bytes:
    """Encode a JSON-like value using viser's hybrid placeholder/binary layout."""
    buffers: list[memoryview] = []
    payload = _to_hybrid_value(document, buffers)
    if not isinstance(payload, dict):
        raise TypeError("Hybrid document root must encode to a mapping.")
    if _BINARY_LENGTHS_KEY in payload:
        raise ValueError(f"Reserved key {_BINARY_LENGTHS_KEY!r} is not allowed.")
    if buffers:
        payload = {
            **payload,
            _BINARY_LENGTHS_KEY: [len(buffer) for buffer in buffers],
        }

    packed = msgspec.msgpack.encode(payload)
    out = bytearray()
    out.extend(len(packed).to_bytes(8, "little"))
    out.extend(packed)
    for buffer in buffers:
        out.extend(b"\x00" * (-len(out) % 8))
        out.extend(buffer)
    return bytes(out)


def serialize_zstd_hybrid_document(
    document: object,
    *,
    level: int,
) -> bytes:
    hybrid = encode_hybrid_document(document)
    compressed = zstandard.ZstdCompressor(level=level).compress(hybrid)
    return len(hybrid).to_bytes(8, "little") + compressed



from __future__ import annotations

import dataclasses
import uuid
from typing import Literal

from typing_extensions import override

from . import _viser_private as impl

RuntimeMethod = Literal[
    "applyMessageUpdate",
    "configure",
    "evictBlock",
    "loadBlock",
    "pause",
    "play",
    "refresh",
    "seek",
    "setSpeed",
]


@dataclasses.dataclass
class Viser4dRuntimeMessage(impl.Message):
    method: RuntimeMethod
    payload: object | None

    @override
    def redundancy_key(self) -> str:
        return str(uuid.uuid4())


@dataclasses.dataclass
class Viser4dRuntimeReadyMessage(impl.Message):
    @override
    def redundancy_key(self) -> str:
        return type(self).__name__
